<?php
declare(strict_types = 1);

/**
 * Gestion de la géolocalisation.
 *
 * @license http://www.gnu.org/licenses/gpl.html
 * @link http://www.igalerie.org/
 */
class Geolocation
{
	/**
	 * Vérifie et retourne des coordonnées au format décimal.
	 *
	 * @param mixed $latitude
	 * @param mixed $longitude
	 *
	 * @return array
	 */
	public static function checkCoordinates($latitude, $longitude): array
	{
		if (($latitude = self::getDecimalCoordinate($latitude)) === FALSE
		 || ($longitude = self::getDecimalCoordinate($longitude)) === FALSE)
		{
			return [];
		}

		if ($latitude > 90 || $latitude < -90 || $longitude > 180 || $longitude < -180)
		{
			return [];
		}

		return
		[
			'latitude' => substr((string) $latitude, 0, 17),
			'longitude' => substr((string) $longitude, 0, 18)
		];
	}

	/**
	 * Modifie les coordonnées de plusieurs catégories.
	 *
	 * @param array $cat_coords
	 *   Coordonnées associées à l'identifiant d'une catégorie.
	 *   Exemple :
	 *   $cat_coords = [7 => ['latitude' => 46.5, 'longitude' => 3]];
	 *
	 * @return int
	 *   Retourne le nombre de catégories affectées
	 *   ou -1 en cas d'erreur.
	 */
	public static function editCategories(array $cat_coords): int
	{
		return self::_edit($cat_coords, 'cat', 'categories');
	}

	/**
	 * Modifie les coordonnées de plusieurs fichiers.
	 *
	 * @param array $items_coords
	 *   Coordonnées associées à l'identifiant d'un fichier.
	 *
	 * @return int
	 *   Retourne le nombre de fichiers affectés
	 *   ou -1 en cas d'erreur.
	 */
	public static function editItems(array $items_coords): int
	{
		return self::_edit($items_coords, 'item', 'items');
	}

	/**
	 * Convertit une coordonnée (latitude ou longitude) depuis
	 * un format sexagésimal (45° 54' 36") ou décimal (45,91°),
	 * en format décimal (45,91).
	 *
	 * @param mixed $coord
	 *
	 * @return mixed
	 *   Retourne une chaîne contenant la coordonnée au format décimal, ou
	 *   FALSE si la coordonnée n'est pas fournie dans un format valide.
	 */
	public static function getDecimalCoordinate($coord)
	{
		// (float) 1.0E-12 => (string) 0.000000000001
		// (float) 7.45101E-11 => (string) 0.0000000000745101
		$to_string = function($val): string
		{
			$val = strval($val);
			if (preg_match('`^(\d+(?:\.\d+))E\-(\d+)$`i', $val, $m))
			{
				$val = '0.' . str_repeat('0', $m[2] - 1) . str_replace('.', '', $m[1]);
			}
			return $val;
		};

		$coord = $to_string($coord);

		if (strlen($coord) > 99)
		{
			return FALSE;
		}

		$coord = Utility::trimAll(str_replace(',', '.', $coord));
		$coord = preg_replace('`\s+`', ' ', $coord);

		// Format décimal.
		if (preg_match('`^[-+]?\d{1,3}(?:\.\d+)?\s*(?:°)?(?:\s*[ENOSW])?$`i', $coord))
		{
			return $to_string((float) $coord);
		}

		// Format sexagésimal.
		$degree = '\s*(\d{1,3})\s*(?:°)?';
		$minute = '(?:\s*(\d{1,2})\s*[\'′]?)?';
		$second = '(?:\s*(?<!°\s)(?<![°\d])(\d+(?:\.\d+)?)\s*["″]?)?';
		$regexp = '([-+])?' . $degree . $minute . $second . '\s*(?:°\s*)?([ENOSW])?';
		if (preg_match('`^' . $regexp . '$`iu', $coord, $m))
		{
			$s = isset($m[5]) ? (in_array(strtoupper($m[5]), ['E','N']) ? '' : '-') : $m[1];
			return $to_string($s . ($m[2] + (($m[3] ?? 0) / 60) + (($m[4] ?? 0) / 3600)));
		}

		return FALSE;
	}



	/**
	 * Modifie les coordonnées de plusieurs fichiers ou catégories.
	 *
	 * @param array $coords
	 * @param string $prefix
	 * @param string $table
	 *
	 * @return int
	 *   Retourne le nombre d'objets affectés
	 *   ou -1 en cas d'erreur.
	 */
	private static function _edit(array $coords, string $prefix, string $table): int
	{
		$params = [];

		// Vérification du format des coordonnées.
		foreach ($coords as $id => &$c)
		{
			if (!isset($c['latitude']) || !isset($c['longitude']))
			{
				continue;
			}

			if (Utility::trimAll($c['latitude']) === ''
			 && Utility::trimAll($c['longitude']) === '')
			{
				$params[$id] =
				[
					'lat' => NULL,
					'long' => NULL,
					'id' => (int) $id
				];
			}

			else if ($latlng = self::checkCoordinates($c['latitude'], $c['longitude']))
			{
				$params[$id] =
				[
					'lat' => $latlng['latitude'],
					'long' => $latlng['longitude'],
					'id' => (int) $id
				];
			}
		}
		if (!$params)
		{
			return 0;
		}

		// On ne conserve que les objets dont les coordonnées sont différentes.
		$sql = 'SELECT %1$s_id, %1$s_lat, %1$s_long FROM {%2$s} WHERE %1$s_id IN (%3$s)';
		if (!DB::execute(sprintf($sql, $prefix, $table, DB::inInt(array_column($params, 'id')))))
		{
			return -1;
		}
		foreach (DB::fetchAll($prefix . '_id') as $id => &$i)
		{
			if ($i[$prefix . '_lat'] == $params[$id]['lat']
			 && $i[$prefix . '_long'] == $params[$id]['long'])
			{
				unset($params[$id]);
			}
		}
		if (!$params)
		{
			return 0;
		}

		// Mise à jour de la base de données.
		$sql = 'UPDATE {%2$s} SET %1$s_lat = :lat, %1$s_long = :long WHERE %1$s_id = :id';
		if (!DB::execute(sprintf($sql, $prefix, $table), array_values($params)))
		{
			return -1;
		}

		return DB::rowCount();
	}
}
?>